class ImageConfig: ...
class StatsContext: ...

def gaussian_2d(width, height, h, k, a, b): ...
def gaussian_1d(width, height, h, k, ystd, xwidth): ...
def scaling2pix(x, y, bounds): ...
def coords2pix(x, y, bounds): ...
def write_image(data, filename, config, context, axis_limits) -> None: ...
def dress_image(img, bounds): ...
def points2img(fname) -> None: ...



import hdcms_bindings
import re
import os
import sys
import numpy as np

def file_ok(fname):
    """ensure that fname is a file"""
    if not (os.path.isfile(fname) and os.access(fname, os.R_OK)):
        raise RuntimeError(f"File {fname} doesn't exist or isn't readable")

def filenames2stats1d(filenames, start, end, num_bins, scaling):
    """take filenames and convert them into a 1d summary statistic"""
    for f in filenames.split(","):
        file_ok(f)
    return hdcms_bindings.filenames_to_stats_1d(filenames, start=start, end=end, num_bins=num_bins, scaling=scaling)

def filenames2stats2d(filenames, scaling, xtol):
    """take filenames and convert them into a 2d summary statistic"""
    for f in filenames.split(","):
        if not (os.path.isfile(f) and os.access(f, os.R_OK)):
            raise RuntimeError(f"File {f} doesn't exist or isn't readable")
    return hdcms_bindings.filenames_to_stats_2d(filenames, scaling=scaling, xtol=xtol)

def regex2filenames(regex, dir="."):
    """takes regex, finds all files that match and convert them into list of filenames"""
    files = [f for f in os.listdir(dir) if os.path.isfile(os.path.join(dir, f))]

    r = re.compile(regex)
    matches = []
    for f in files:
        match = r.match(f)
        if match:
            matches.append(f)

    if len(matches) == 0:
        raise RuntimeError(f"No matches for {regex} in directory {dir}")

    full_paths = list(map(lambda f: os.path.join(dir, f), matches))
    return ','.join(full_paths)

def regex2stats1d(regex, dir=".", start=0, end=900, num_bins=9000, scaling='m'):
    """takes regex, converts list of filenames that match into 1d summary stat"""
    filenames = regex2filenames(regex, dir)
    return filenames2stats1d(filenames, start, end, num_bins, scaling)

def regex2stats2d(regex, dir=".", scaling='m', xtol=float('inf')):
    """takes regex, converts list of filenames that match into 2d summary stat
       example: regex2stats2d(r"CM1_2_\\d.txt", dir="../data")"""
    filenames = regex2filenames(regex, dir)
    return filenames2stats2d(filenames, scaling, xtol)

def file2filenames(filename):
    """takes file with filenames on separate lines and joins them into a string of filenames separated by commas"""
    with open(filename) as f:
        return ",".join(f.readlines())

def file2stats1d(filename, start=0, end=900, num_bins=9000, scaling='m'):
    """takes file with filenames on separate lines converts them into a 1d summary statistic"""
    return filenames2stats1d(file2filenames(filename), start, end, num_bins, scaling)

def file2stats2d(filename, scaling='m', xtol=float('inf')):
    """takes file with filenames on separate lines converts them into a 2d summary statistic
       example: file2stats2d("./compound1_high_res.txt")"""
    return filenames2stats2d(file2filenames(filename), scaling, xtol)

def get_unique_tmpdir(name="hdcms-numpy-tmp"):
    """creates a unique temporary directory, in unix it's /tmp, on windows its C:\\Users\\AppData\\Local\\Temp"""
    dir = f"/tmp" if sys.platform != "win32" else f"C:\\Users\\AppData\\Local\\Temp"
    _, existing_dirs, _ = next(os.walk(dir))
    while name in existing_dirs:
        name += "0"
    return os.path.join(dir, name)

def array2stats1d(*args, start=0, end=900, num_bins=9000, scaling='m'):
    """takes a varargs list of numpy arrays and converts them into 1d summary statistic"""
    dir = get_unique_tmpdir()
    lst = []
    for i, arr in args:
        filename = os.path.join(dir, f"{i}-numpy-array.txt")
        np.savetxt(filename, arr)
        lst.append(filename)
    return filenames2stats1d(",".join(lst), start, end, num_bins, scaling)

def array2stats2d(*args, scaling='m', xtol=float('inf')):
    """takes a varargs list of numpy arrays and converts them into 2d summary statistic"""
    dir = get_unique_tmpdir()
    lst = []
    for i, arr in args:
        filename = os.path.join(dir, f"{i}-numpy-array.txt")
        np.savetxt(filename, arr)
        lst.append(filename)
    return filenames2stats2d(",".join(lst), scaling, xtol)

# figures out the comparison function needed and checks that the input is valid size
def compare(*args, npeaks=None):
    """takes varargs list of summary statistic and computes the similarity according to its arguments,
       so if all arrays are 1d summary stats it will compare them using 1d similarity, same with 2d, if
       anything doesn't match it will raise a helpful error message"""
    is_using_2d = (npeaks != None)
    if len(args) <= 1:
        raise RuntimeError("Must compare at least 2 summary statistics")

    len_of_dim_2 = args[0].shape[1]

    # make sure it is either 2 or 4
    if len_of_dim_2 not in [2,4]:
        raise RuntimeError(f"Incorrect dimension: recieved {args[0].shape} must be (_, 2) or (_, 4)")

    # make sure they are using the right number of peaks
    if len_of_dim_2 == 2 and is_using_2d:
        raise RuntimeError("Mismatch dimension: supplied npeaks={npeaks} but using an array with dimensions {args[0].shape}. To use 2d comparison must be (_, 4)")
    elif len_of_dim_2 == 4:
        is_using_2d = True

    # verify they all have the same length of second dimension
    for arr in args:
        if arr.shape[1] != len_of_dim_2:
            raise RuntimeError(f"Mismatch dimension: recieved {arr.shape} and {args[0].shape}, they must be the same and either (_, 2) or (_, 4)")

    if len(args) == 2:
        if is_using_2d:
            return hdcms_bindings.compare_compound_2d(args[0], args[1])
        else:
            return hdcms_bindings.compare_compound_1d(args[0], args[1])
    else:
        if is_using_2d:
            return hdcms_bindings.compare_all_2d(args)
        else:
            return hdcms_bindings.compare_all_1d(args)




import scipy.stats
import numpy as np
from matplotlib import pyplot as plt

SCALE_RADIUS = 0.01
ADD_RADIUS = 0.001
MEAN = 200
N = 100

def data(dist):
    """create data"""
    sample = dist.rvs(size=(N, 2), loc=MEAN) # we get garbage for y values
    sample[:,1] = dist.pdf(sample[:,0], loc=MEAN)
    return smear(sample)

def plot(data, title="no title provided"):
    """plot data on a graph"""
    plt.scatter(data[:,0], data[:,1])
    plt.title(title)
    plt.xlim(190, 210)
    plt.ylim(0,0.6)
    plt.show()

def smear(data):
    """add noise to data"""
    scale = 1 + np.random.random() * SCALE_RADIUS - SCALE_RADIUS
    add = np.random.random() * ADD_RADIUS - ADD_RADIUS
    data[:,1] *= scale
    data[:,1] += add
    return data

def generate_examples(visualize=False):
    """use scipy statistics distributions to generate synthetic data"""
    for i in range(10):
        with open(f"gaus_{i}.txt", "w") as f:
            d = data(scipy.stats.norm)
            for pt in d:
                print(pt[0], pt[1], file=f)
            if visualize:
                plot(d, "gaussian")

    for i in range(10):
        with open(f"laplace_{i}.txt", "w") as f:
            d = data(scipy.stats.laplace)
            for pt in d:
                print(pt[0], pt[1], file=f)
            if visualize:
                plot(d, "laplace")

    for i in range(10):
        with open(f"cauchy_{i}.txt", "w") as f:
            d = data(scipy.stats.cauchy)
            for pt in d:
                print(pt[0], pt[1], file=f)
            if visualize:
                plot(d, "cauchy")

    for i in range(3):
        with open(f"unknown_{i}.txt", "w") as f:
            d = data(scipy.stats.laplace)
            for pt in d:
                print(pt[0], pt[1], file=f)
            if visualize:
                plot(d, "unknown (laplace)")

    with open("params.txt", "w") as f:
        print("Parameters when we last ran `gen.py`:", file=f)
        print(f"SCALE_RADIUS: {SCALE_RADIUS}", file=f)
        print(f"ADD_RADIUS: {ADD_RADIUS}", file=f)
        print(f"MEAN: {MEAN}", file=f)
        print(f"N: {N}", file=f)




SCALE_RADIUS: float
ADD_RADIUS: float
MEAN: int
N: int

def data(dist): ...
def plot(data, title: str = ...) -> None: ...
def smear(data): ...
def generate_examples(visualize: bool = ...) -> None: ...



from _typeshed import Incomplete

def file_ok(fname) -> None: ...
def filenames2stats1d(filenames): ...
def filenames2stats2d(filenames): ...
def regex2filenames(regex, dir: str = ...): ...
def regex2stats1d(regex, dir: str = ...): ...
def regex2stats2d(regex, dir: str = ...): ...
def file2filenames(filename): ...
def file2stats1d(filename): ...
def file2stats2d(filename): ...
def get_unique_tmpdir(name: str = ...): ...
def array2stats1d(*args): ...
def array2stats2d(*args): ...
def compare(*args, npeaks: Incomplete | None = ...): ...



from .gen import generate_examples as generate_examples
from .stats import array2stats1d as array2stats1d, array2stats2d as array2stats2d, compare as compare, file2stats1d as file2stats1d, file2stats2d as file2stats2d, filenames2stats1d as filenames2stats1d, filenames2stats2d as filenames2stats2d, regex2stats1d as regex2stats1d, regex2stats2d as regex2stats2d
from .visualize import write_image as write_image
from .verify import is_valid_ms_data_format as is_valid_ms_data_format



# https://stackoverflow.com/questions/49829783/draw-a-gradual-change-ellipse-in-skimage

import numpy as np
import sys
import matplotlib.pyplot as plt
from PIL import Image

class ImageConfig:
    """this class is used to keep track of all the parameters of the image
    """
    def __init__(self, size=(3600, 1200), std_scale=1, desingularization=1e-5, axis_thickness=3, xlabel="m/z values", ylabel="intensities", title="hdc visualization"):
        self.xpixels, self.ypixels = size

        # how much to scale the stddev
        self.std_scale = std_scale
        # how much to add to the stddev after scale
        self.desingularization = desingularization

        # thickness of axes
        self.axis_thickness = axis_thickness

        self.xlabel = xlabel
        self.ylabel = ylabel
        self.title = title

class StatsContext:
    """this class is used by 1d code to keep track of the bin parameters
    """
    def __init__(self, start=0, stop=899.9, num_bins=9000):
        self.start = start
        self.stop = stop
        self.num_bins = num_bins
        self.bin_width = (stop - start) / (num_bins - 1)

def gaussian_2d(width, height, h, k, a, b):
    """ goal: output an array of pixels with 2d gaussian drawn on
    recall: a 2d gaussian is a mean and standard deviation of a peak

    width and height are from the image config [refactor: pass the config itself and use field access]
    (h,k) are the center of the gaussian
    (a, b) are the standard deviations of the gaussian
    """
    # Generate (x,y) coordinate arrays
    y,x = np.mgrid[-k:height-k,-h:width-h] 
    # returns an array [[[-k, -k+1, ..., height-k-1, height-k], ...], 
    #                   [[-k, ..., -k], ..., [height-k, ..., height-k]]]

    weights = np.exp((-1/2) * ((x / a)**2 + ((y / b)**2)))
    if np.max(weights):
        return weights / np.max(weights)
    else:
        return weights

def gaussian_1d(h, k, ystd, xwidth, config):
    """ goal: output an array of pixels with 1d gaussian drawn on
    recall: a 1d gaussian is a mean and standard deviation of a bin

    (h,k) are the center of the gaussian
    ystd is the standard deviation of the gaussian
    xwidth is how wide the bin is
    config is the image config (size, etc)
    """
    # don't want to go off screen
    lowest_x_rendered = max(int(h - (xwidth/2)), 0)
    highest_x_rendered = min(int(h + (xwidth/2)), config.xpixels)
    if lowest_x_rendered == highest_x_rendered:
        print(f"pixels {(h,k)} w ystd: {ystd}, bin_width: {xwidth} in image of size {(config.xpixels, config.ypixels)} (width, height), ", file=sys.stderr)
        return np.zeros((config.ypixels, config.xpixels))
    y,_ = np.mgrid[-k:config.ypixels-k, lowest_x_rendered:highest_x_rendered]
    # so sets y to [[-k, -k, -k, ..., -k, -k],
    #               [-(k-1), -(k-1), -(k-1), ..., -(k-1)],
    #               ...
    #               [(height-k-1), (height-k-1), ..., (height-k-1)]]
    # with the number of columns equal to (highest_x_rendered - lowest_x_rendered)
    weights = np.exp((-1/2) * ((y / ystd)**2))

    ret = np.zeros((config.ypixels, config.xpixels))
    ret[:, lowest_x_rendered:highest_x_rendered] = weights
    if weights.size != 0 and np.max(weights) != 0:
        return ret / np.max(weights)
    else:
        return ret

def coordinate_rectangle_to_pixels(rect, axis_limits, config):
    """ goal: convert (x,y,width,height) from coordinates to pixel
    note: we need to flip the y values
    """
    x, y, width, height = rect
    xmin, ymin, xmax, ymax = axis_limits
    xscale = config.xpixels / (xmax - xmin)
    yscale = config.ypixels / (ymax - ymin)

    x = (x - xmin)*xscale # we don't add anything bc min pixel value is 0
    y = (y - ymin)*yscale # we don't add anything bc min pixel value is 0
    return (x, config.ypixels - y, width * xscale, height * yscale)

def write_image(data, color=(1, 0, 0), config=ImageConfig(), context=StatsContext(), axis_limits=None):
    data = np.array(data)
    if len(data[0]) == 4:
        data[:,2] = data[:,2] * config.std_scale + config.desingularization
        data[:,3] = data[:,3] * config.std_scale + config.desingularization
    elif len(data[0]) == 2:
        data[:,1] = data[:,1] * config.std_scale + config.desingularization
    else:
        raise RuntimeError(f"Incorrect dimension: recieved {data.shape} must be nx2 or nx4")

    # We need to auto adjust the axis limits if not provided
    if not axis_limits:
        yborder = 0.1
        xborder = 0.1

        if len(data[0]) == 4:
            width = np.max(data[:, 0]) - np.min(data[:, 0])
            height = np.max(data[:, 1]) - np.min(data[:, 1])
        else:
            width = (np.max(np.nonzero(data[:, 0])) + 1) * context.bin_width
            height = np.max(data[:, 0])
        # min_x, min_y, max_x, max_y
        axis_limits = (width * (-xborder), height * (-yborder), width * (1+xborder), height * (1+yborder))

    if len(data[0]) == 4:
        assert(np.min(data[:, 1]) >= 0 and np.min(data[:,0]) >= 0)

    img = np.ones((config.ypixels, config.xpixels, 3))

    for i, peak in enumerate(data):
        if len(peak) == 4:
            h, k, a, b = coordinate_rectangle_to_pixels(peak, axis_limits, config)
            red, green, blue = color
            img[:, :, 0] -= (1-red) * gaussian_2d(config.xpixels, config.ypixels, h, k, a, b)
            img[:, :, 1] -= (1-green) * gaussian_2d(config.xpixels, config.ypixels, h, k, a, b)
            img[:, :, 2] -= (1-blue) * gaussian_2d(config.xpixels, config.ypixels, h, k, a, b)
        else:
            if np.sum(peak[0]) != 0.:
                rect_center_x = context.start + i*context.bin_width + context.bin_width/2
                h, k, width, sd = coordinate_rectangle_to_pixels((rect_center_x, peak[0], context.bin_width, peak[1]), axis_limits, config)
                red, green, blue = color
                img[:, :, 0] -= (1-red) * gaussian_1d(h, k, sd, width, config)
                img[:, :, 1] -= (1-green) * gaussian_1d(h, k, sd, width, config)
                img[:, :, 2] -= (1-blue) * gaussian_1d(h, k, sd, width, config)

    img = dress_image(img, axis_limits, config)
    return Image.fromarray(img)

def dress_image(img, axis_limits, config):
    """this function takes the straight output of the gaussians and adds axes, tickmarks, text
    """
    # convert to bytes 0..1 -> 0..255
    img = np.uint8(img * 255)

    xmin, ymin, xmax, ymax = axis_limits
    xlen = img.shape[1]
    ylen = img.shape[0]
    old_figure_num = plt.gcf().number # figure to restore
    figsize = plt.rcParams["figure.figsize"][0] # we take x value
    APPROX_MARGIN_RATIO_OF_IMG = 0.3 # correct for the fact that plt tick marks and axes take up part of image
    fig, ax = plt.subplots(dpi=(1/APPROX_MARGIN_RATIO_OF_IMG)*ylen/figsize)
    ax.imshow(img, extent=(xmin, xmax, ymin, ymax))
    ax.set_aspect(((xmax-xmin)/(ymax-ymin))/(xlen/ylen))
    ax.set_title(config.title)
    ax.set_xlabel(config.xlabel)
    ax.set_ylabel(config.ylabel)

    fig.canvas.draw()
    buf = fig.canvas.tostring_rgb()
    ncols, nrows = fig.canvas.get_width_height()
    img = np.frombuffer(buf, dtype=np.uint8)
    # note: the nrows and ncols order
    img = img.reshape(nrows, ncols, 3)

    # reactivate old figure
    plt.figure(old_figure_num)
    return img



from .stats import regex2stats1d, regex2stats2d
from .stats import array2stats1d, array2stats2d
from .stats import file2stats1d, file2stats2d
from .stats import filenames2stats1d, filenames2stats2d
from .stats import compare
from .visualize import write_image
from .gen import generate_examples
from .verify import is_valid_ms_data_format



import re

def isfloat(num):
    try:
        float(num)
        return True
    except ValueError:
        return False

def assert_all_floats(l):
    for ele in l:
        if not isfloat(ele):
            raise RuntimeError(f"found: nonfloat {ele}")

def is_valid_ms_data_format(path_name):
    with open(path_name) as f:
        line_iter = iter(f.readlines())

        line = None

        # process header
        for line in line_iter:
            if line[0] != '#':
                break

        if not line:
            raise RuntimeError(f"empty file")

        fields = re.split('[ \t,]+', line)
        assert_all_floats(fields)
        if len(fields) != 2:
            raise RuntimeError("got too many values per line")

        for line in line_iter:
            fields = re.split('[ \t,]+', line)
            assert_all_floats(fields)
            if len(fields) != 2:
                raise RuntimeError("got too many values per line")

